from abc import abstractmethod
from typing import Any, Dict, Generic, Iterable, List, Optional, Type, TypeVar

from pydantic import BaseModel as PydanticBaseModel

T = TypeVar("T")


class Field(PydanticBaseModel):
    """A custom field for a QIF object.

    Parameters
    ----------
    line_code : str
        The line code for the custom field. This is the first character that
        appears on the line in the QIF file.
    attr : str
        The attribute name for the custom field. The value of the custom field
        will be stored in this attribute and will be accessible via dot
        notation.
    type : type
        The type of the custom field. This is the type that the value of the
        custom field will be converted to.
    """

    line_code: str
    attr: str
    type: Type

    def __eq__(self, other) -> bool:
        return self.line_code == other.line_code

    # Include __lt__ method so that the Field objects can be sorted
    def __lt__(self, other) -> bool:
        return (
            len(self.line_code) < len(other.line_code)
            or self.line_code < other.line_code
        )


class BaseModel(PydanticBaseModel, Generic[T]):
    class Config:
        extra = "allow"

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    @classmethod
    def add_custom_field(
        cls,
        line_code: str,
        attr: str,
        field_type: Type,
    ) -> None:
        """Add a custom field to the class for parsing from a QIF.

        Parameters
        ----------
        line_code : str
            The line code of the extra field. This is the first character that
            appears on the line in the QIF file.
        attr : str
            The attribute name of the extra field. The value of the extra field
            will be stored in this attribute and will be accessible via dot
            notation.
        field_type : str
            The type of the extra field. This is the type that the value of the
            extra field will be converted to.
        """
        lst = getattr(cls, "__CUSTOM_FIELDS", []).copy()

        new_field = Field(line_code=line_code, attr=attr, type=field_type)
        if new_field in lst:
            # Overwrite existing field if it exists
            lst = [field if field != new_field else new_field for field in lst]
        else:
            lst.append(new_field)

        setattr(cls, "__CUSTOM_FIELDS", lst)

    @classmethod
    def _get_custom_fields(cls) -> List[Field]:
        """Return a list of the custom fields for the class, reverse ordered by
        line code length."""
        return sorted(getattr(cls, "__CUSTOM_FIELDS", []), reverse=True)

    @classmethod
    @abstractmethod
    def from_list(cls, lst: List[str]) -> T:
        pass

    @classmethod
    def from_string(cls, string: str, separator: str = "\n") -> T:
        """Create a class instance from a string."""
        return cls.from_list(string.split(separator))

    def to_dict(self, ignore: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        """Convert the class instance to a dict."""
        if ignore is None:
            ignore = []
        return self.dict(exclude=set(ignore))



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, List, Optional

from quiffen import utils
from quiffen.core.base import BaseModel, Field
from quiffen.core.class_type import Class
from quiffen.core.transaction import TransactionLike, TransactionList


class AccountType(str, Enum):
    """An enum representing the different account types allowed by QIF."""

    CASH = "Cash"
    BANK = "Bank"
    CREDIT_CARD = "CCard"
    OTH_A = "Oth A"
    OTH_L = "Oth L"
    INVOICE = "Invoice"
    INVST = "Invst"


class Account(BaseModel):
    """
    A class representing a QIF account.

    Examples
    --------
    Creating an Account instance, then adding a transaction.

    >>> import quiffen
    >>> from datetime import datetime
    >>> acc = quiffen.Account(name='Example name', desc='Some description')
    >>> acc
    Account(name='Example name', desc='Some description')
    >>> tr = quiffen.Transaction(date=datetime.now(), amount=150.0)
    >>> acc.set_header(quiffen.AccountType.BANK)
    >>> acc.add_transaction(tr)
    >>> acc.transactions
    {
        'Bank': [
            Transaction(
                date=datetime.datetime(2021, 7, 2, 18, 31, 47, 817025),
                amount=150.0,
            ),
        ],
    }

    Creating an account instance from a section list in a QIF file.

    >>> import quiffen
    >>> string = (
    ...     '!Account\\nNPersonal Bank Account\\n'
    ...     'DMy Personal bank account with Barclays.\\n^\\n'
    ... )
    >>> acc = quiffen.Account.from_string(string)
    >>> acc
    Account(
        name='Personal Bank Account',
        desc='My Personal bank account with Barclays.',
        ...,
    )
    """

    name: str
    desc: Optional[str] = None
    account_type: Optional[AccountType] = None
    credit_limit: Optional[Decimal] = None
    balance: Optional[Decimal] = None
    date_at_balance: Optional[datetime] = None
    transactions: Dict[AccountType, TransactionList] = {}
    _last_header: Optional[AccountType] = None

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __eq__(self, other) -> bool:
        if not isinstance(other, Account):
            return False
        return self.name == other.name and self.account_type == other.account_type

    def __str__(self) -> str:
        properties = ""
        ignore = ["_last_header"]
        for object_property, value in self.__dict__.items():
            if value and object_property not in ignore:
                if object_property == "transactions":
                    num_transactions = sum(len(value[header]) for header in value)
                    properties += f"\n\tTransactions: {num_transactions}"
                elif object_property == "account_type":
                    properties += f"\n\tAccount Type: {value.value}"
                else:
                    properties += (
                        f"\n\t"
                        f'{object_property.replace("_", " ").strip().title()}: '
                        f"{value}"
                    )
        return "Account:" + properties

    def set_header(self, header: AccountType) -> None:
        """Set the last header used.

        Parameters
        ----------
        header : AccountType
            The last header used.
        """
        self._last_header = header

    def add_transaction(
        self,
        transaction: TransactionLike,
        header: Optional[AccountType] = None,
    ) -> None:
        """Add a transaction to the dict of TransactionList objects.

        Parameters
        ----------
        transaction : TransactionLike
            The Transaction-type object to be added.
        header : AccountType
             The header under which the transaction falls. Will be used as a key
             in the dict of transactions.

        Raises
        -------
        RuntimeError
            If there is no header provided, and no ``last_header`` set.
        """
        if not header and not self._last_header:
            raise RuntimeError("No header provided, and no last header set.")

        if not header:
            header = self._last_header
        else:
            try:
                header = AccountType(header.strip().split(":")[-1])
                self._last_header = header
            except ValueError as e:
                raise ValueError("Header must be a valid AccountType.") from e

        if not header:
            raise RuntimeError("No header provided, and no last header set.")

        if header not in self.transactions:
            self.transactions[header] = []

        self.transactions[header].append(transaction)

    def merge(self, other: Account) -> None:
        """Merge another account into this one.

        Parameters
        ----------
        other : Account
            The other account to merge into this one.
        """
        self.desc = self.desc or other.desc
        self.account_type = self.account_type or other.account_type
        self.credit_limit = self.credit_limit or other.credit_limit
        self.balance = self.balance or other.balance
        self.date_at_balance = self.date_at_balance or other.date_at_balance

        for header in other.transactions:
            if header not in self.transactions:
                self.transactions[header] = []
            self.transactions[header].extend(other.transactions[header])

    def to_qif(
        self, date_format: str = "%Y-%m-%d", classes: Optional[Dict[str, Class]] = None
    ) -> str:
        """Return a QIF-formatted string of this account.

        Parameters
        ----------
        date_format : str
            The date format to use for the date field.
        classes : Dict[str, Class]
            A dict of Class objects in the account's transactions, keyed by
            their name.

        Returns
        -------
        str
            A QIF-formatted string of this account.
        """
        qif = "!Account\n"
        qif += f"N{self.name}\n"
        if self.desc:
            qif += f"D{self.desc}\n"
        if self.account_type:
            qif += f"T{self.account_type.value}\n"
        if self.credit_limit:
            qif += f"L{self.credit_limit}\n"
        if self.balance:
            qif += f"${self.balance}\n"
        if self.date_at_balance:
            qif += f"/{self.date_at_balance}\n"

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        for header, header_transactions in self.transactions.items():
            qif += f"^\n!Type:{header.value}\n"
            qif += "^\n".join(
                transaction.to_qif(
                    date_format=date_format,
                    classes=classes,
                )
                for transaction in header_transactions
            )
        return qif

    @classmethod
    def from_list(cls, lst: List[str], day_first: bool = False) -> Account:
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the account.
        day_first : bool, default=False
             Whether the day comes before the month in the date.

        Returns
        -------
        Account
            An Account object created from the QIF strings.
        """
        kwargs: Dict[str, Any] = {}
        for field in lst:
            line_code, field_info = utils.parse_line_code_and_field_info(field)
            if not line_code:
                continue

            # Check if current line is a custom field
            kwargs, found = utils.add_custom_field_to_object_dict(
                field=field,
                custom_fields=cls._get_custom_fields(),
                object_dict=kwargs,
            )
            if found or line_code == "!":
                continue

            if line_code == "N":
                kwargs["name"] = field_info
            elif line_code == "D":
                kwargs["desc"] = field_info
            elif line_code == "T":
                kwargs["account_type"] = field_info
            elif line_code == "L":
                kwargs["credit_limit"] = field_info.replace(",", "")
            elif line_code in {"$", "£"}:
                kwargs["balance"] = field_info.replace(",", "")
            elif line_code == "/":
                balance_date = utils.parse_date(field_info, day_first)
                kwargs["date_at_balance"] = balance_date
            else:
                raise ValueError(f"Unknown line code: {line_code}")

        return cls(**kwargs)



from __future__ import annotations

from typing import Any, Dict, List, Optional

from quiffen import utils
from quiffen.core.base import BaseModel, Field


class Security(BaseModel):
    """A class representing a security in a QIF file

    Parameters
    ----------
    name : str
        The name of the security
    symbol : str
        The symbol of the security
    type : str
        The type of the security
    goal : str
        The goal of the security
    line_number : int
        The line number of the security in the QIF file
    """

    name: Optional[str] = None
    symbol: Optional[str] = None
    type: Optional[str] = None
    goal: Optional[str] = None
    line_number: Optional[int] = None

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __str__(self) -> str:
        return_str = "Security:"
        if self.name:
            return_str += f"\n\tName: {self.name}"
        if self.symbol:
            return_str += f"\n\tSymbol: {self.symbol}"
        if self.type:
            return_str += f"\n\tType: {self.type}"
        if self.goal:
            return_str += f"\n\tGoal: {self.goal}"

        return return_str

    def merge(self, other: Security) -> None:
        """Merge another security into this one.

        Parameters
        ----------
        other : Security
            The other security to merge into this one.
        """
        self.name = self.name or other.name
        self.symbol = self.symbol or other.symbol
        self.type = self.type or other.type
        self.goal = self.goal or other.goal

    def to_qif(self) -> str:
        """Converts a Security to a QIF string"""
        qif = "!Type:Security\n"

        if self.name:
            qif += f"N{self.name}\n"
        if self.symbol:
            qif += f"S{self.symbol}\n"
        if self.type:
            qif += f"T{self.type}\n"
        if self.goal:
            qif += f"G{self.goal}\n"

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        return qif

    @classmethod
    def from_list(
        cls,
        lst: List[str],
        line_number: Optional[int] = None,
    ) -> Security:
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the transaction.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF
            file.

        Returns
        -------
        Security
            A class instance representing the security.
        """
        kwargs: Dict[str, Any] = {}

        for field in lst:
            line_code, field_info = utils.parse_line_code_and_field_info(field)
            if not line_code:
                continue

            # Check if current line is a custom field
            kwargs, found = utils.add_custom_field_to_object_dict(
                field=field,
                custom_fields=cls._get_custom_fields(),
                object_dict=kwargs,
            )
            if found:
                continue

            if line_code == "N":
                kwargs["name"] = field_info
            elif line_code == "S":
                kwargs["symbol"] = field_info
            elif line_code == "T":
                kwargs["type"] = field_info
            elif line_code == "G":
                kwargs["goal"] = field_info
            else:
                raise ValueError(f"Unknown line code: {line_code}")

        if line_number is not None:
            kwargs["line_number"] = line_number

        return cls(**kwargs)

    @classmethod
    def from_string(
        cls,
        string: str,
        separator: str = "\n",
        line_number: Optional[int] = None,
    ) -> Security:
        """Return a class instance from a QIF string.

        Parameters
        ----------
        string : str
            String containing QIF information about the transaction.
        separator : str, default='\n'
            The separator between QIF fields.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF
            file.

        Returns
        -------
        Security
            A class instance representing the security.
        """
        return cls.from_list(
            lst=string.split(separator),
            line_number=line_number,
        )



from __future__ import annotations

import csv
import io
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from pydantic.types import FilePath

from quiffen import utils
from quiffen.core.account import Account, AccountType
from quiffen.core.base import BaseModel, Field
from quiffen.core.category import Category, add_categories_to_container
from quiffen.core.class_type import Class
from quiffen.core.investment import Investment
from quiffen.core.security import Security
from quiffen.core.transaction import Transaction

try:
    import pandas as pd

    PANDAS_INSTALLED = True
except ModuleNotFoundError:
    PANDAS_INSTALLED = False

VALID_TRANSACTION_ACCOUNT_TYPES = [
    "!type:cash",
    "!type:bank",
    "!type:ccard",
    "!type:otha",
    "!type:othl",
    "!type:invoice",
]


class QifDataType(str, Enum):
    """An Enum representing the different types of data that can be found in a
    Qif object."""

    TRANSACTIONS = "transactions"
    INVESTMENTS = "investments"
    CLASSES = "classes"
    CATEGORIES = "categories"
    ACCOUNTS = "accounts"
    SECURITIES = "securities"


class ParserException(Exception):
    pass


class Qif(BaseModel):
    """
    The main class of the package. For parsing QIF files.

    See the readme for usage examples.

    Parameters
    ----------
    accounts : dict, default=None
        A dict of accounts in the form {'Account Name': account_object}.
    categories : dict, default=None
        A dict of categories in the form {'Category Name': category_object}.
    classes : dict, default=None
        A dict of classes in the form {'Class Name': class_object}.
    """

    accounts: Dict[str, Account] = {}
    categories: Dict[str, Category] = {}
    classes: Dict[str, Class] = {}
    securities: Dict[str, Security] = {}

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __str__(self) -> str:
        accounts_str = "\n".join(str(acc) for acc in self.accounts.values())
        categories_str = "\n".join(str(cat) for cat in self.categories.values())
        classes_str = "\n".join(str(cls) for cls in self.classes.values())
        securities_str = "\n".join(str(sec) for sec in self.securities.values())

        if not (accounts_str or categories_str or classes_str):
            return "Empty Qif object"

        return_str = "QIF\n===\n\n"

        if accounts_str:
            return_str += f"Accounts\n--------\n\n{accounts_str}\n\n"
        if categories_str:
            return_str += f"Categories\n----------\n\n{categories_str}\n\n"
        if classes_str:
            return_str += f"Classes\n-------\n\n{classes_str}\n\n"
        if securities_str:
            return_str += f"Securities\n----------\n\n{securities_str}\n\n"

        return return_str

    @classmethod
    def from_list(cls, lst: List[str]) -> Qif:
        raise NotImplementedError(
            "This method is not implemented for Qif objects. Use Qif.parse to "
            "parse a QIF file."
        )

    @classmethod
    def parse(
        cls,
        path: Union[FilePath, str],
        separator: str = "\n",
        day_first: bool = False,
        encoding: str = "utf-8",
    ) -> Qif:
        """Return a class instance from a QIF file.

        Parameters
        ----------
        path : Union[FilePath, str]
            The path to the QIF file.
        separator : str, default='\n'
            The line separator for the QIF file. This probably won't need
            changing.
        day_first : bool, default=False
            Whether the day or month comes first in the date.
        encoding : str, default='utf-8'
            The encoding of the QIF file.

        Returns
        -------
        Qif
            A Qif object containing all the data in the QIF file.
        """
        path = Path(path)
        if path.suffix.lower() != ".qif":
            raise ParserException("The file must be a QIF file.")

        if not path.exists():
            raise ParserException("The file does not exist.")

        data = path.read_text(encoding=encoding).strip().strip("\n")

        if not data:
            raise ParserException("The file is empty.")

        accounts: Dict[str, Account] = {}
        last_account = None
        categories: Dict[str, Category] = {}
        classes: Dict[str, Class] = {}
        securities: Dict[str, Security] = {}

        sections = data.split("^")
        last_header = None
        line_number = 1

        for section in sections:
            if not section:
                continue

            section_lines = section.split(separator)
            if not section_lines:
                continue

            # Allow for comments and blank lines at the top of sections
            for i, line in enumerate(section_lines):
                header_line = line
                if line.strip() and line[0] != "#":
                    line_number += i
                    section_lines = section_lines[i:]
                    break
            else:
                # Empty section
                continue

            if header_line[0] != "!":
                if not last_header:
                    raise ParserException(
                        f"Line {line_number}: " f"No header found before transactions."
                    )
                header_line = last_header

            last_header = header_line

            sanitised_section_lines = [
                line
                for line in section_lines
                if line.strip() and line.strip()[0] != "!"
            ]

            if not sanitised_section_lines:
                continue

            # Check for new categories and accounts first, otherwise it's a
            # transaction so a default account is created
            if "!Type:Cat" in header_line:
                # Section contains category information
                new_category = Category.from_list(sanitised_section_lines)
                categories = add_categories_to_container(  # type: ignore
                    new_category,
                    categories,
                )
            elif "!Type:Class" in header_line:
                new_class = Class.from_list(sanitised_section_lines)
                if new_class.name in classes:
                    classes[new_class.name].merge(new_class)
                else:
                    classes[new_class.name] = new_class
            elif "!Account" in header_line:
                new_account = Account.from_list(sanitised_section_lines)
                if new_account.name in accounts:
                    accounts[new_account.name].merge(new_account)
                else:
                    accounts[new_account.name] = new_account
                last_account = new_account.name
            elif "!Type:Invst" in header_line:
                # Investment
                new_investment = Investment.from_list(
                    sanitised_section_lines,
                    day_first=day_first,
                    line_number=line_number,
                )

                if last_account is None:
                    raise ParserException(
                        f"Line {line_number}: "
                        "No account found before investment. "
                        "This should not happen."
                    )

                accounts[last_account].add_transaction(
                    new_investment,
                    AccountType("Invst"),
                )
            elif "!Type:Security" in header_line:
                # Security
                new_security = Security.from_list(
                    sanitised_section_lines,
                    line_number=line_number,
                )
                if new_security.symbol is None:
                    raise ParserException(
                        f"Line {line_number}: " f"No symbol found for security."
                    )
                securities[new_security.symbol] = new_security
            elif "!Type" in header_line and not accounts:
                # Accounts is empty and there's a transaction, so create default
                # account to put transactions in
                default_account = Account(
                    name="Quiffen Default Account",
                    desc=(
                        "The default account created by Quiffen when no other "
                        "accounts were present"
                    ),
                )
                accounts[default_account.name] = default_account
                last_account = default_account.name

            if header_line.lower().replace(" ", "") in (
                VALID_TRANSACTION_ACCOUNT_TYPES
            ):
                # Other transaction type
                new_transaction, new_classes = Transaction.from_list(
                    sanitised_section_lines,
                    day_first=day_first,
                    line_number=line_number,
                )

                if last_account is None:
                    raise ParserException(
                        f"Line {line_number}: "
                        "No account found before transactions. "
                        "This should not happen."
                    )

                accounts[last_account].add_transaction(
                    new_transaction,
                    AccountType(header_line.split(":")[1].strip()),
                )

                if new_transaction.category:
                    root = new_transaction.category.traverse_up()[-1]
                    if root.name in categories:
                        categories[root.name].merge(root)
                    else:
                        categories[root.name] = root

                for class_name, new_class in new_classes.items():
                    if class_name in classes:
                        classes[class_name].merge(new_class)
                    else:
                        classes[class_name] = new_class

            line_number += len(section.split("\n"))

        return cls(
            accounts=accounts,
            categories=categories,
            classes=classes,
            securities=securities,
        )

    def add_account(self, new_account: Account) -> None:
        """Add a new account to the Qif object"""
        if new_account.name in self.accounts:
            self.accounts[new_account.name].merge(new_account)
        else:
            self.accounts[new_account.name] = new_account

    def remove_account(self, account_name: str) -> Account:
        """Remove an account from this Qif object"""
        try:
            return self.accounts.pop(account_name)
        except KeyError as e:
            raise KeyError(
                f'Account "{account_name}" does not exist in this Qif object.'
            ) from e

    def add_category(self, new_category: Category) -> None:
        """Add a new category to the Qif object"""
        self.categories = add_categories_to_container(  # type: ignore
            new_category,
            self.categories,
        )

    def remove_category(
        self,
        category_name: str,
        keep_children: bool = True,
    ) -> Category:
        """Remove a category from this Qif object"""
        try:
            category = self.categories.pop(category_name)
        except KeyError as e:
            raise KeyError(
                f'Category "{category_name}" does not exist in this Qif object.'
            ) from e

        if keep_children:
            print("Keeping children")
            for child in category.children:
                child.set_parent(None)
                self.add_category(child)

        return category

    def add_class(self, new_class: Class) -> None:
        """Add a new class to the Qif object"""
        if new_class.name in self.classes:
            self.classes[new_class.name].merge(new_class)
        else:
            self.classes[new_class.name] = new_class

    def remove_class(self, class_name: str) -> Class:
        """Remove a class from this Qif object"""
        try:
            return self.classes.pop(class_name)
        except KeyError as e:
            raise KeyError(
                f'Class "{class_name}" does not exist in this Qif object.'
            ) from e

    def add_security(self, new_security: Security) -> None:
        """Add a new security to the Qif object"""
        if not new_security.symbol:
            raise ValueError(
                "Cannot add a security without a symbol to the Qif object."
            )

        if new_security.symbol in self.securities:
            self.securities[new_security.symbol].merge(new_security)
        else:
            self.securities[new_security.symbol] = new_security

    def remove_security(self, security_symbol: str) -> Security:
        """Remove a security from this Qif object"""
        try:
            return self.securities.pop(security_symbol)
        except KeyError as e:
            raise KeyError(
                f'Security "{security_symbol}" does not exist in this Qif ' f"object."
            ) from e

    def to_qif(
        self,
        path: Optional[Union[FilePath, str, None]] = None,
        date_format: str = "%Y-%m-%d",
        encoding: str = "utf-8",
    ) -> str:
        """Convert the Qif object to a QIF file"""
        qif = ""

        if self.categories:
            qif += (
                "^\n".join(category.to_qif() for category in self.categories.values())
                + "^\n"
            )

        if self.classes:
            qif += "^\n".join(cls.to_qif() for cls in self.classes.values()) + "^\n"

        if self.accounts:
            qif += (
                "^\n".join(
                    account.to_qif(date_format=date_format, classes=self.classes)
                    for account in self.accounts.values()
                )
                + "^\n"
            )

        if path:
            Path(path).write_text(qif, encoding=encoding)

        return qif

    def _get_data_dicts(
        self,
        data_type: QifDataType = QifDataType.TRANSACTIONS,
        date_format: Optional[str] = "%Y-%m-%d",
        ignore: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """Converts specified data from the Qif object to a list of dicts"""
        if ignore is None:
            ignore = []

        if data_type == QifDataType.TRANSACTIONS:
            data_dicts = []
            for account in self.accounts.values():
                for transaction_list in account.transactions.values():
                    data_dicts += [
                        transaction.to_dict(ignore=ignore)
                        for transaction in transaction_list
                        if not isinstance(transaction, Investment)
                    ]
        elif data_type == QifDataType.INVESTMENTS:
            data_dicts = []
            for account in self.accounts.values():
                for transaction_list in account.transactions.values():
                    data_dicts += [
                        transaction.to_dict(ignore=ignore)
                        for transaction in transaction_list
                        if isinstance(transaction, Investment)
                    ]
        elif data_type == QifDataType.ACCOUNTS:
            data_dicts = [
                account.to_dict(ignore=ignore) for account in self.accounts.values()
            ]
        elif data_type == QifDataType.CATEGORIES:
            data_dicts = [
                category.to_dict(ignore=ignore) for category in self.categories.values()
            ]
        elif data_type == QifDataType.CLASSES:
            data_dicts = [
                class_.to_dict(ignore=ignore) for class_ in self.classes.values()
            ]
        elif data_type == QifDataType.SECURITIES:
            data_dicts = [
                security.to_dict(ignore=ignore) for security in self.securities.values()
            ]
        else:
            raise ValueError(
                f"Invalid data_type: {data_type}. Must be one of "
                f"{list(QifDataType)}"
            )

        # Format and hide private fields
        return [  # type: ignore
            utils.apply_csv_formatting_to_container(
                {
                    k: v
                    for k, v in data_dict.items()
                    if not k.startswith("_") and k not in ignore
                },
                date_format=date_format,
            )
            for data_dict in data_dicts
        ]

    def to_csv(
        self,
        path: Optional[Union[FilePath, str, None]] = None,
        data_type: QifDataType = QifDataType.TRANSACTIONS,
        date_format: str = "%Y-%m-%d",
        ignore: Optional[List[str]] = None,
        delimiter: str = ",",
        quote_character: str = '"',
        encoding: str = "utf-8",
    ) -> str:
        """Convert part of the Qif object to a CSV file. The data_type
        parameter can be used to specify which part of the Qif object to
        convert to CSV. The default is to convert transactions.

        Parameters
        ----------
        path : Union[FilePath, str, None], optional
            The path to write the CSV file to, by default None
        data_type : QifDataType, optional
            The type of data to convert to CSV, by default
            QifDataType.TRANSACTIONS
        date_format : str, optional
            The date format to use when converting dates to strings, by
            default '%Y-%m-%d'
        ignore : List[str], optional
            A list of fields to ignore when converting to CSV, by default
            None
        delimiter : str, optional
            The delimiter to use when writing the CSV file, by default ','
        quote_character : str, optional
            The quote character to use when writing the CSV file, by
            default '"'
        encoding : str, default='utf-8'
            The encoding to use when writing the CSV file
        """
        if ignore is None:
            ignore = []

        data_dicts = self._get_data_dicts(
            data_type=data_type,
            date_format=date_format,
            ignore=ignore,
        )

        headers: Set[str] = set()
        for data_dict in data_dicts:
            headers.update(k for k in data_dict.keys())

        output = io.StringIO()

        writer = csv.DictWriter(
            output,
            fieldnames=headers,
            extrasaction="ignore",  # Ignore extra fields (e.g. private fields)
            dialect="unix",  # Use Unix line endings
            delimiter=delimiter,
            quotechar=quote_character,
        )
        writer.writeheader()
        for data_dict in data_dicts:
            writer.writerow(data_dict)

        return_value = output.getvalue()

        if path:
            Path(path).write_text(return_value, encoding=encoding)

        return return_value

    def to_dataframe(
        self,
        data_type: QifDataType = QifDataType.TRANSACTIONS,
        ignore: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Convert part of the Qif object to a Pandas DataFrame. The
        data_type parameter can be used to specify which part of the Qif
        object to convert to a DataFrame. The default is to convert
        transactions.

        Parameters
        ----------
        data_type : QifDataType, optional
            The type of data to convert to a DataFrame, by default
            QifDataType.TRANSACTIONS
        ignore : List[str], optional
            A list of fields to ignore when converting to a DataFrame, by
            default None

        Returns
        -------
        pd.DataFrame
            The data as a Pandas DataFrame
        """
        if not PANDAS_INSTALLED:
            raise ModuleNotFoundError(
                "The pandas module must be installed to use this method"
            )

        if ignore is None:
            ignore = []

        data_dicts = self._get_data_dicts(
            data_type=data_type,
            date_format=None,
            ignore=ignore,
        )

        return pd.DataFrame(data_dicts)



from __future__ import annotations

import logging
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import root_validator, validator

from quiffen import utils
from quiffen.core.base import BaseModel, Field
from quiffen.core.category import (
    Category,
    add_categories_to_container,
    create_categories_from_hierarchy,
)
from quiffen.core.class_type import Class
from quiffen.core.investment import Investment
from quiffen.core.split import Split

logger = logging.getLogger(__name__)


class Transaction(BaseModel):
    """
    A class used to represent a transaction.

    Parameters
    ----------
    date : datetime
        Date transaction occurred. May or may not include timestamp.
    amount : decimal.Decimal
        The amount of the transaction. May be positive or negative.
    memo : str, default=None
        Also known as the reference. A string describing the purpose behind the
        transaction.
    cleared : str, default=None
        The cleared status of this transaction. See the QIF standards for valid
        values.
    payee : str, default=None
        The name of the payee on the other end of the transaction. The payee is
        the receiver of the money if amount is negative, else they are the
        sender of the money.
    payee_address : str, default=None
        The address of the aforementioned payee.
    category : Category, default=None
        The category object that represents the transaction.
    check_number : int, default=None
        The check number if the transaction relates to a check
    reimbursable_expense : bool, default=None
        Whether this transaction is flagged as a reimbursable business expense.
    small_business_expense : bool, default=None
        Whether this transaction is flagged as a small business expense.
    to_account : str, default=None
        The account the transaction was sent to, if applicable.
    first_payment_date : datetime.datetime, default=None
        If this transaction was completed over multiple days, the first payment
        date.
    loan_length : decimal.Decimal, default=None
        The length of the loan, if applicable.
    num_payments : int, default=None
        If this payment was split over multiple payments, the number of such
        payments.
    periods_per_annum : int, default=None
        The periods per annum for this transaction.
    interest_rate : decimal.Decimal, default=None
        The interest rate on this transaction.
    current_loan_balance : decimal.Decimal, default=None
        The current loan balance, if applicable.
    original_loan_amount : decimal.Decimal, default=None
        The original loan amount, if applicable.
    line_number : int, default=None
        The line number of the header line of the transaction in the QIF file.
    splits : TransactionList, default=None
        If this transaction has multiple categories (e.g. an Amazon purchase of
        an electrical item and a book), it can be split in QIF files to
        represent this. Each split has its own memo, category and amount. This
        parameter is a TransactionList containing the splits for this
        transaction.

    Examples
    --------
    Create a Transaction instance, then convert to a dict, ignoring the date.

    >>> import quiffen
    >>> import decimal
    >>> from datetime import datetime
    >>> cat = quiffen.Category('Finances')
    >>> tr = quiffen.Transaction(
    ...     date=datetime.now(),
    ...     amount=decimal.Decimal(150.60),
    ...     category=cat,
    ... )
    >>> tr
    Transaction(
        date=datetime.datetime(2021, 7, 5, 10, 45, 40, 48195),
        amount=150.6,
        category=Category(name='Finances',expense=True, hierarchy='Finances'),
    )
    >>> print(tr)
    Transaction:
        Date: 2021-07-05 10:45:40.048195
        Amount: 150.6
        Category: Finances
    >>> tr.to_dict(ignore=['date'], dictify_category=True)
    {
        'amount': 150.6,
        'category': {
            'name': 'Finances',
            'expense': True,
            'income': False,
            'hierarchy': 'Finances',
            'children': [],
        },
    }
    """

    date: datetime
    amount: Decimal
    memo: Optional[str] = None
    cleared: Optional[str] = None
    payee: Optional[str] = None
    payee_address: Optional[str] = None
    category: Optional[Category] = None
    check_number: Optional[Union[int, str]] = None
    reimbursable_expense: Optional[bool] = None
    small_business_expense: Optional[bool] = None
    to_account: Optional[str] = None
    first_payment_date: Optional[datetime] = None
    loan_length: Optional[Decimal] = None
    num_payments: Optional[int] = None
    periods_per_annum: Optional[int] = None
    interest_rate: Optional[Decimal] = None
    current_loan_balance: Optional[Decimal] = None
    original_loan_amount: Optional[Decimal] = None
    line_number: Optional[int] = None
    splits: List[Split] = []
    _split_categories: Dict[str, Category] = {}
    _last_split: Optional[Split] = None

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __str__(self) -> str:
        properties = ""
        ignore = ["_last_split", "_is_split"]
        for object_property, value in self.__dict__.items():
            if value and object_property not in ignore:
                if object_property == "category":
                    properties += f"\n\tCategory: {value.name}"
                elif object_property == "_split_categories":
                    properties += f"\n\tSplit Categories: {list(value.keys())}"
                elif object_property == "splits":
                    properties += f"\n\tSplits: {len(value)}"
                else:
                    properties += (
                        f"\n\t"
                        f'{object_property.replace("_", " ").strip().title()}: '
                        f"{value}"
                    )

        return "Transaction:" + properties

    @root_validator(pre=True)
    def create_split_categories(cls, values: Dict[str, Any]) -> Dict:
        if splits := values.get("splits"):
            for split in splits:
                if split.category:
                    values["_split_categories"] = add_categories_to_container(
                        split.category,
                        values.get("_split_categories", {}),
                    )
            values["_last_split"] = splits[-1]
        return values

    @validator("splits")
    def check_split_percentages_and_amounts(
        cls,
        splits: List[Split],
        values: Dict[str, Any],
    ) -> List[Split]:
        total_percent = sum(split.percent for split in splits if split.percent)
        total_amount = sum(split.amount for split in splits if split.amount is not None)
        if total_percent - 100 > 0.01:
            raise ValueError("Split percentages cannot exceed 100% of the transaction")
        if abs(total_amount) - abs(values.get("amount", 0)) > 0.01:
            raise ValueError(
                "Split amounts cannot exceed the amount of the transaction"
            )
        return splits

    @property
    def split_categories(self) -> Dict[str, Category]:
        return self._split_categories

    @property
    def is_split(self) -> bool:
        """Returns True if the transaction is a split transaction, False
        otherwise."""
        return len(self.splits) > 0

    def add_split(self, split: Split) -> None:
        """Add a Split to Transaction."""
        if (
            split.percent
            and sum(s.percent for s in self.splits if s.percent is not None)
            + split.percent
            - 100
            > 0.01
        ):
            raise ValueError(
                "The sum of the split percentages cannot be greater than 100."
            )

        if split.amount:
            abs_sum_of_splits = abs(
                sum(s.amount for s in self.splits if s.amount is not None)
                + split.amount
            )
            if abs_sum_of_splits - abs(self.amount) > 0.01:
                raise ValueError(
                    "The sum of the split amounts cannot be greater than the "
                    "transaction amount."
                )

        self.splits.append(split)

    def remove_splits(self, **filters) -> List[Split]:
        """Remove splits from Transaction if they match a set of filters

        Takes filters as kwargs and removes splits that match all filters. If
        no filters are provided, all splits are removed.
        """
        if not filters:
            current_splits = self.splits.copy()
            self.splits = []
            return current_splits

        to_remove = []
        for split in self.splits:
            if all(getattr(split, attr) == value for attr, value in filters.items()):
                to_remove.append(split)

        self.splits = [split for split in self.splits if split not in to_remove]

        return to_remove

    @staticmethod
    def _create_class_from_category_string(
        category_string: str,
        classes: Dict[str, Class],
    ) -> Tuple[Optional[str], str, Dict[str, Class]]:
        if "/" in category_string:
            field_info, class_name = category_string.split("/")
            if class_name not in classes:
                classes[class_name] = Class(name=class_name)
            return class_name, field_info, classes
        return None, category_string, classes

    def to_qif(
        self,
        date_format: str = "%Y-%m-%d",
        classes: Optional[Dict[str, Class]] = None,
    ) -> str:
        """Converts a Transaction to a QIF string"""
        if classes is None:
            classes = {}

        qif = f"D{self.date.strftime(date_format)}\n"
        qif += f"T{self.amount}\n"
        if self.memo:
            qif += f"M{self.memo}\n"
        if self.cleared:
            qif += f"C{self.cleared}\n"
        if self.payee:
            qif += f"P{self.payee}\n"
        if self.payee_address:
            qif += f"A{self.payee_address}\n"
        if self.category:
            parent_class = None
            for cls in classes.values():
                for category in cls.categories:
                    if category.find_child(self.category.name):
                        parent_class = cls
                        break

            qif += f"L{self.category.hierarchy}"
            if parent_class:
                qif += f"/{parent_class.name}"
            qif += "\n"
        if self.check_number:
            qif += f"N{self.check_number}\n"
        if self.reimbursable_expense:
            qif += f"F{self.reimbursable_expense}\n"
        if self.to_account:
            qif += f"L[{self.to_account}]\n"
        if self.first_payment_date:
            qif += f"1{self.first_payment_date.strftime(date_format)}\n"
        if self.loan_length:
            qif += f"2{self.loan_length}\n"
        if self.num_payments:
            qif += f"3{self.num_payments}\n"
        if self.periods_per_annum:
            qif += f"4{self.periods_per_annum}\n"
        if self.interest_rate:
            qif += f"5{self.interest_rate}\n"
        if self.current_loan_balance:
            qif += f"6{self.current_loan_balance}\n"
        if self.original_loan_amount:
            qif += f"7{self.original_loan_amount}\n"
        if self.splits:
            for split in self.splits:
                qif += split.to_qif(date_format=date_format, classes=classes)

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        return qif

    @classmethod
    def from_list(
        cls,
        lst: List[str],
        day_first: bool = False,
        line_number: Optional[int] = None,
    ) -> Tuple[Transaction, Dict[str, Class]]:
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the transaction.
        day_first : bool, default=False
             Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF
            file.

        Returns
        -------
        Transaction
            A Transaction object created from the QIF strings.
        Dict[str, Class]
            A dictionary of Class objects created from the QIF strings.
        """
        kwargs: Dict[str, Any] = {}
        classes: Dict[str, Class] = {}
        splits: List[Split] = []
        current_split: Optional[Split] = None

        for field in lst:
            line_code, field_info = utils.parse_line_code_and_field_info(field)
            if not line_code:
                continue

            # Check if current line is a custom field
            kwargs, found = utils.add_custom_field_to_object_dict(
                field=field,
                custom_fields=cls._get_custom_fields(),
                object_dict=kwargs,
            )
            if found:
                continue

            if line_code == "S":
                _, field_info, classes = cls._create_class_from_category_string(
                    field_info,
                    classes,
                )

                split_category = create_categories_from_hierarchy(field_info)
                new_split = Split(category=split_category)
                splits.append(new_split)
                current_split = new_split
            elif line_code == "D":
                transaction_date = utils.parse_date(field_info, day_first)
                if not splits:
                    kwargs["date"] = transaction_date
                elif current_split:
                    current_split.date = transaction_date
            elif line_code == "E":
                if current_split is None:
                    logger.warning(
                        f"No split yet given for memo '{field_info}', skipping"
                    )
                else:
                    current_split.memo = field_info
            elif line_code in {"$", "£"}:
                if current_split:
                    current_split.amount = utils.parse_decimal(field_info)
            elif line_code == "%":
                if current_split:
                    current_split.percent = utils.parse_decimal(
                        field_info.split(" ")[0]
                    )
            elif line_code in {"T", "U"}:
                amount = utils.parse_decimal(field_info)
                if not splits:
                    kwargs["amount"] = amount
                elif current_split:
                    current_split.amount = amount
            elif line_code == "M":
                if not splits:
                    kwargs["memo"] = field_info
                elif current_split:
                    current_split.memo = field_info
            elif line_code == "C":
                if not splits:
                    kwargs["cleared"] = field_info
                elif current_split:
                    current_split.cleared = field_info
            elif line_code == "P":
                kwargs["payee"] = field_info
            elif line_code == "A":
                if not splits:
                    kwargs["payee_address"] = field_info
                elif current_split:
                    current_split.payee_address = field_info
            elif line_code == "L":
                (
                    class_name,
                    field_info,
                    classes,
                ) = cls._create_class_from_category_string(
                    field_info,
                    classes,
                )

                # 'L' can represent both categories and the 'to' transfer
                # account. Transfer accounts are denoted by [ ]
                if field_info.startswith("["):
                    if not splits:
                        kwargs["to_account"] = field_info[1:-1]
                    elif current_split:
                        current_split.to_account = field_info[1:-1]
                else:
                    category = create_categories_from_hierarchy(field_info)
                    category_root = category.traverse_up()[-1]
                    if not splits:
                        # If there's already a category, add the new category
                        # as a child
                        if "category" in kwargs:
                            category_root.set_parent(kwargs["category"])
                        kwargs["category"] = category
                    elif current_split:
                        category_root.set_parent(current_split.category)
                        current_split.category = category

                    if class_name:
                        classes[class_name].add_category(category)
            elif line_code == "N":
                if not splits:
                    kwargs["check_number"] = field_info
                elif current_split:
                    current_split.check_number = field_info
            elif line_code == "F":
                kwargs["reimbursable_expense"] = field_info or True
            elif line_code == "1":
                kwargs["first_payment_date"] = utils.parse_date(
                    field_info,
                    day_first,
                )
            elif line_code == "2":
                kwargs["loan_length"] = field_info.replace(",", "")
            elif line_code == "3":
                kwargs["num_payments"] = field_info.replace(",", "")
            elif line_code == "4":
                kwargs["periods_per_annum"] = field_info.replace(",", "")
            elif line_code == "5":
                kwargs["interest_rate"] = field_info.replace(",", "")
            elif line_code == "6":
                kwargs["current_loan_balance"] = field_info.replace(",", "")
            elif line_code == "7":
                kwargs["original_loan_amount"] = field_info.replace(",", "")
            else:
                raise ValueError(f"Unknown line code: {line_code}")

        if line_number is not None:
            kwargs["line_number"] = line_number

        # Set splits percentage if they don't already have one
        total = utils.parse_decimal(kwargs.get("amount", 0))
        if splits and total:
            for split in splits:
                if split.percent is None and split.amount is not None:
                    split.percent = utils.parse_decimal(
                        round(split.amount / total * 100, 2)
                    )
                # Check if the split percentage is correct
                elif (
                    split.percent is not None
                    and split.amount is not None
                    and not (
                        utils.parse_decimal(round(split.percent, 2))
                        == utils.parse_decimal(
                            round(
                                split.amount / total * 100,
                                2,
                            )
                        )
                    )
                ):
                    logger.warning(
                        f"Split percentage ({split.percent}) does not match "
                        f"the split amount ({split.amount}) for transaction "
                        f"on line {line_number}."
                    )
        elif splits and not total:
            for split in splits:
                split.percent = None

        kwargs["splits"] = splits
        return cls(**kwargs), classes

    @classmethod
    def from_string(
        cls,
        string: str,
        separator: str = "\n",
        day_first: bool = False,
        line_number: Optional[int] = None,
    ) -> Tuple[Transaction, Dict[str, Class]]:
        """Return a class instance from a QIF string.

        Parameters
        ----------
        string : str
            String containing QIF information about the transaction.
        separator : str, default='\n'
            The separator between QIF fields.
        day_first : bool, default=False
             Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF
            file.

        Returns
        -------
        Transaction
            A Transaction object created from the QIF strings.
        Dict[str, Class]
            A dictionary of classes created from the QIF strings.
        """
        lines = string.split(separator)
        return cls.from_list(lines, day_first, line_number)


TransactionLike = Union[Transaction, Investment, Split]
TransactionList = List[TransactionLike]



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, List, Optional

from quiffen import utils
from quiffen.core.base import BaseModel, Field


class Investment(BaseModel):
    """A class used to represent an investment.

    Acts almost identically to the Transaction class, but with a few different
    parameters and no splits.

    Parameters
    ----------
    date : datetime
        Date investment occurred. May or may not include timestamp.
    action : str, default=None
        The investment action (Buy, Sell, etc.)
    security : str, default=None
        The security name.
    price : Decimal, default=None
        The price of the security.
    quantity : Decimal, default=None
        The quantity of the security bought, sold, etc.
    cleared : str, default=None
        The cleared status of this investment. See the QIF standards for valid
        values.
    amount : Decimal, default=None
        The overall amount of this investment.
    memo : str, default=None
        Also known as the reference. A string describing the purpose behind the
        investment.
    first_line : str, default=None
        The first line of the investment.
    to_account : str, default=None
        The to account of the investment, if applicable.
    transfer_amount : Decimal, default=None
        The amount transferred for the investment.
    commission : Decimal, default=None
        The commission paid/received on the investment.
    line_number : int, default=None
        The line number of the investment in the QIF file.
    """

    date: datetime
    action: Optional[str] = None
    security: Optional[str] = None
    price: Optional[Decimal] = None
    quantity: Optional[Decimal] = None
    cleared: Optional[str] = None
    amount: Optional[Decimal] = None
    memo: Optional[str] = None
    first_line: Optional[str] = None
    to_account: Optional[str] = None
    transfer_amount: Optional[Decimal] = None
    commission: Optional[Decimal] = None
    line_number: Optional[int] = None

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __str__(self) -> str:
        properties = ""
        for object_property, value in self.__dict__.items():
            if value:
                properties += (
                    f"\n\t"
                    f'{object_property.replace("_", " ").strip().title()}: '
                    f"{value}"
                )

        return "Investment:" + properties

    def to_qif(
        self,
        date_format: str = "%Y-%m-%d",
        **_,  # To keep the same signature as Transaction.to_qif
    ) -> str:
        """Converts an Investment to a QIF string"""
        qif = f"D{self.date.strftime(date_format)}\n"

        if self.action:
            qif += f"N{self.action}\n"
        if self.security:
            qif += f"Y{self.security}\n"
        if self.price:
            qif += f"I{self.price}\n"
        if self.quantity:
            qif += f"Q{self.quantity}\n"
        if self.cleared:
            qif += f"C{self.cleared}\n"
        if self.amount:
            qif += f"T{self.amount}\n"
        if self.memo:
            qif += f"M{self.memo}\n"
        if self.first_line:
            qif += f"P{self.first_line}\n"
        if self.to_account:
            qif += f"L{self.to_account}\n"
        if self.transfer_amount:
            qif += f"${self.transfer_amount}\n"
        if self.commission:
            qif += f"O{self.commission}\n"

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        return qif

    @classmethod
    def from_list(
        cls,
        lst: List[str],
        day_first: bool = False,
        line_number: Optional[int] = None,
    ) -> Investment:
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the investment.
        day_first : bool, default=False
            Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the investment in the QIF
            file.

        Returns
        -------
        Investment
            An Investment object created from the QIF strings.
        """
        kwargs: Dict[str, Any] = {}
        for field in lst:
            line_code, field_info = utils.parse_line_code_and_field_info(field)
            if not line_code:
                continue

            # Check if current line is a custom field
            kwargs, found = utils.add_custom_field_to_object_dict(
                field=field,
                custom_fields=cls._get_custom_fields(),
                object_dict=kwargs,
            )
            if found:
                continue

            # Check the QIF line code for banking-related operations, then
            # append to kwargs.
            if line_code == "D":
                transaction_date = utils.parse_date(field_info, day_first)
                kwargs["date"] = transaction_date
            elif line_code == "N":
                kwargs["action"] = field_info
            elif line_code == "Y":
                kwargs["security"] = field_info
            elif line_code == "I":
                kwargs["price"] = field_info.replace(",", "")
            elif line_code == "Q":
                kwargs["quantity"] = field_info.replace(",", "")
            elif line_code == "C":
                kwargs["cleared"] = field_info
            elif line_code in {"T", "U"}:
                kwargs["amount"] = field_info.replace(",", "")
            elif line_code == "M":
                kwargs["memo"] = field_info
            elif line_code == "P":
                kwargs["first_line"] = field_info
            elif line_code == "L":
                kwargs["to_account"] = field_info
            elif line_code == "$":
                kwargs["transfer_amount"] = field_info.replace(",", "")
            elif line_code == "O":
                kwargs["commission"] = field_info.replace(",", "")
            else:
                raise ValueError(f"Unknown line code: {line_code}")

        if line_number is not None:
            kwargs["line_number"] = line_number

        return cls(**kwargs)

    @classmethod
    def from_string(
        cls,
        string: str,
        separator: str = "\n",
        day_first: bool = False,
        line_number: Optional[int] = None,
    ) -> Investment:
        """Return a class instance from a QIF string.

        Parameters
        ----------
        string : str
            String containing QIF information about the investment.
        separator : str, default='\n'
            The separator between QIF fields.
        day_first : bool, default=False
            Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the investment in the QIF
            file.

        Returns
        -------
        Investment
            An Investment object created from the QIF string.
        """
        lst = string.split(separator)
        return cls.from_list(lst, day_first, line_number)



from __future__ import annotations

from decimal import Decimal
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Union

from pydantic import validator

from quiffen import utils
from quiffen.core.base import BaseModel, Field


class CategoryType(str, Enum):
    """Enum representing the different types of categories in a QIF file."""

    EXPENSE = "expense"
    INCOME = "income"


class Category(BaseModel):
    """
    A node-like class used to represent a category. Can be built into trees to
    represent category families.

    Parameters
    ----------
    name : str
        The name of the category.
    desc : str, default=None
        The category's description.
    tax_related : bool, default=None
        Whether the category represents a tax related group of transactions.
    category_type : CategoryType, default='expense'
        The type of category. Can be either 'expense' or 'income'.
    budget_amount : decimal.Decimal, default=None
        The budget amount for this category.
    tax_schedule_info : str, default=None
        Information about the tax schedule for this category.
    children : list of Category
        A list of the category's children objects.
    parent : Category, default=None
        The parent category for this category.
    hierarchy : str, default=None
        The category hierarchy, separated by : characters.

    Examples
    --------
    Creating a category tree, then rendering it to console.

    >>> import quiffen
    >>> food = quiffen.Category(name='Food')
    >>> food
    Category(name='Food', expense=True, hierarchy='Food')
    >>> essentials = quiffen.Category(name='Essentials')
    >>> food.add_child(essentials)
    >>> pastas = quiffen.Category(name='Pastas')
    >>> essentials.add_child(pastas)
    >>> pastas.hierarchy
    'Food:Essentials:Pastas'
    >>> meat = quiffen.Category(name='Meat')
    >>> food.add_child(meat)
    >>> print(food.render_tree())
    Food (root)
    └─ Essentials
       └─ Pastas
    └─ Meat

    Removing a category from a tree by passing in the Category instance or the
    category name string.

    >>> print(food.render_tree())
    Food (root)
    └─ Essentials
       └─ Pastas
    └─ Meat
       └─ Chicken
    >>> meat
    Category(
        name='Meat',
        expense=True,
        parent=Category(name='Food', expense=True, hierarchy='Food'),
        hierarchy='Food:Meat',
    )
    >>> food.remove_child(meat, keep_children=True)
    >>> print(food.render_tree())
    Food (root)
    └─ Essentials
       └─ Pastas
    └─ Chicken
    >>> food.remove_child('Essentials')
    >>> print(food.render_tree())
    Food (root)
    └─ Chicken
    """

    name: str
    desc: Optional[str] = None
    tax_related: Optional[bool] = None
    category_type: CategoryType = CategoryType.EXPENSE
    budget_amount: Optional[Decimal] = None
    tax_schedule_info: Optional[str] = None
    hierarchy: Optional[str] = None
    children: List[Category] = []
    parent: Optional[Category] = None

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __str__(self) -> str:
        properties = ""
        for object_property, value in self.__dict__.items():
            if value:
                if object_property == "parent":
                    properties += (
                        f'\n\tParent: {self.parent.name if self.parent else "None"}'
                    )
                elif object_property == "children":
                    properties += f"\n\tChildren: {len(self.children)}"
                elif object_property == "category_type":
                    properties += f"\n\tCategory Type: {value.value}"
                else:
                    properties += (
                        f"\n\t"
                        f'{object_property.replace("_", " ").strip().title()}: '
                        f"{value}"
                    )
        return "Category:" + properties

    def __lt__(self, other) -> bool:
        return self.name < other.name

    @validator("hierarchy", pre=True, always=True)
    def _set_hierarchy(cls, v: str, values) -> str:
        if not v:
            return values["name"]

        if values.get("parent", None) and v != values["name"]:
            raise ValueError("Hierarchy must match name if no parent is set.")

        if not v.endswith(values["name"]):
            raise ValueError("Hierarchy must end with name.", v, values["name"])

        return v

    def _refresh_hierarchy(self) -> None:
        """Refreshes the hierarchy of the current category and all its children
        recursively."""
        for child in self.children:
            child.hierarchy = (
                self.hierarchy + ":" + child.name if self.hierarchy else child.name
            )
            child._refresh_hierarchy()

    def dict(self, exclude: Optional[Iterable[str]] = None, **_) -> Dict[str, Any]:
        """Return a representation of the Category object as a dict.

        Overwrites pydantic.BaseModel.dict or else it will recurse infinitely
        when trying to serialize the parent and children attributes.

        Parameters
        ----------
        exclude : list of str
            A list of the object's attributes that should not be included in the
            resulting dict.
        """
        if exclude is None:
            exclude = []

        res = {
            key.strip("_"): value
            for (key, value) in self.__dict__.items()
            if key.strip("_") not in exclude
        }

        if self.children and "children" not in exclude:
            res["children"] = [category.name for category in self.children]

        if self.parent and "parent" not in exclude:
            res["parent"] = self.parent.name

        return res

    # This is kept for backwards compatibility
    def to_dict(
        self, ignore: Optional[Iterable[str]] = None, **kwargs
    ) -> Dict[str, Any]:
        """Return a representation of the Category object as a dict.

        Parameters
        ----------
        ignore : iterable of str
            A list of the object's attributes that should not be included in the
            resulting dict.
        """
        return self.dict(exclude=ignore, **kwargs)

    def traverse_down(self) -> List[Category]:
        """Return a flat list of all children, grandchildren etc. of the
        current category.

        The list is ordered starting with the current category, then its
        children, then the children's children etc.
        """
        nodes_to_visit = [self]
        all_children = []
        while nodes_to_visit:
            current_node = nodes_to_visit.pop()
            all_children.append(current_node)
            nodes_to_visit.extend(current_node.children)
        return all_children

    def traverse_up(self: Category) -> List[Category]:
        """Return a list of all parents, grandparents etc. of the current
        category.

        The list is ordered from the current category to the root category.
        """
        current_node = self
        all_parents = [self]
        while current_node.parent:
            all_parents.append(current_node.parent)
            current_node = current_node.parent
        return all_parents

    def find_child(self, node_name: str) -> Union[Category, None]:
        """Returns a node with a given name from the children of current node.

        Will return None if no node with the given name is found. If the node
        being searched for is this node, it will return itself.

        Raises
        ------
        KeyError
            If the category cannot be found.
        """
        nodes_to_visit = [self]
        while nodes_to_visit:
            current_node = nodes_to_visit.pop()
            if current_node.name == node_name:
                return current_node
            nodes_to_visit.extend(current_node.children)
        return None

    def set_parent(self, parent: Union[Category, None]) -> None:
        """Sets the parent of the current category. Set to None to make this
        category a root category.

        Raises
        ------
        ValueError
            If the parent argument is this category.
        """
        if parent is self:
            raise ValueError("Cannot set parent to self.")

        if self.parent and self in self.parent.children:
            self.parent.children.remove(self)

        self.parent = parent

        if parent:
            parent.children.append(self)
            parent._refresh_hierarchy()
        else:
            self.hierarchy = self.name
            self._refresh_hierarchy()

    def add_child(self, child: Union[Category, str]) -> Category:
        """Adds a child category to the current category. Pass a string to
        create a new category with that name.

        Returns
        -------
        Category
            The newly added child category.
        """
        if isinstance(child, str):
            child = Category(name=child)
        child.set_parent(self)
        return child

    def set_children(self, children: Iterable[Category]) -> None:
        """Sets the children of the current category.

        Will remove all existing children and replace them with the new ones.
        """
        self.children = []
        for child in children:
            self.add_child(child)

    def remove_child(
        self,
        child: Union[Category, str],
        keep_children: bool = False,
    ) -> Category:
        """Remove a child category from the current category's hierarchy.

        Parameters
        ----------
        child : str or Category
            The Category object to be removed as a child or the string name of
            the category.
        keep_children : bool, default=False
            Whether the children of the removed category should be kept in the
            tree and moved up a level.

        Returns
        -------
        Category
            The removed category.

        Raises
        ------
        KeyError
            If the category cannot be found.
        """
        if isinstance(child, str):
            child_category = self.find_child(child)
        else:
            child_category = self.find_child(child.name)

        if not child_category:
            raise KeyError(f"Category '{child}' not found.")

        parent = child_category.parent

        if parent:
            new_children = [c for c in parent.children if c != child_category]
        else:
            new_children = []

        if keep_children:
            for grandchild in child_category.children:
                new_children.append(grandchild)

        if parent:
            parent.set_children(new_children)
        child_category.set_parent(None)
        return child_category

    def merge(self, other: Category) -> bool:
        """Recursively merge another category tree into this one.

        Will check through the entire tree for categories with the same name
        and merge them. If a category with the same name is found, the
        children of the other category will be added to the current category
        and the other category will be removed.

        Returns
        -------
        bool
            True if the merge was successful, False otherwise.
        """
        if other.name != self.name:
            for child in self.children:
                success = child.merge(other)
                if success:
                    return True
            return False
        else:
            for other_child in other.children:
                if this_child := self.find_child(other_child.name):
                    this_child.merge(other_child)
                else:
                    self.add_child(other_child)
            return True

    def render_tree(self, _level: int = 0) -> str:
        """Renders a tree-like structure for categories.

        Runs recursively, and uses ``_level`` to keep track of the indentation.
        """
        if not self.children:
            return self.name

        if self.parent is None:
            is_root_str = " (root)"
        else:
            is_root_str = ""

        return (
            self.name
            + f"{is_root_str}\n"
            + "\n".join(
                [
                    "   " * _level + "└─ " + child.render_tree(_level + 1)
                    for child in self.children
                ]
            )
        )

    def _to_qif_string(self) -> str:
        """Return a QIF representation of the individual Category object."""
        qif = f"!Type:Cat\nN{self.hierarchy}\n"

        if self.desc:
            qif += f"D{self.desc}\n"
        if self.tax_related is not None:
            qif += f"T{self.tax_related}\n"

        if self.category_type == CategoryType.INCOME:
            qif += "I\n"
        elif self.category_type == CategoryType.EXPENSE:
            qif += "E\n"

        if self.budget_amount:
            qif += f"B{self.budget_amount}\n"
        if self.tax_schedule_info:
            qif += f"R{self.tax_schedule_info}\n"

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        return qif

    def to_qif(self) -> str:
        """Return a QIF representation of all the categories in the tree."""
        return "^\n".join([c._to_qif_string() for c in self.traverse_down()])

    @classmethod
    def from_list(cls, lst: List[str]) -> Category:
        """Return a Category instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the category.
        """
        kwargs: Dict[str, Any] = {}
        new_parent: Optional[Union[Category, None]] = None
        for field in lst:
            line_code, field_info = utils.parse_line_code_and_field_info(field)
            if not line_code:
                continue

            # Check if current line is a custom field
            kwargs, found = utils.add_custom_field_to_object_dict(
                field=field,
                custom_fields=cls._get_custom_fields(),
                object_dict=kwargs,
            )
            if found:
                continue

            if line_code == "N":
                temp_cat = create_categories_from_hierarchy(field_info)
                kwargs["name"] = temp_cat.name
                kwargs["hierarchy"] = temp_cat.hierarchy
                new_parent = temp_cat.parent
                if new_parent:
                    new_parent.remove_child(temp_cat)
            elif line_code == "D":
                kwargs["desc"] = field_info
            elif line_code == "T":
                kwargs["tax_related"] = field_info or True
            elif line_code == "E":
                kwargs["category_type"] = CategoryType.EXPENSE
            elif line_code == "I":
                kwargs["category_type"] = CategoryType.INCOME
            elif line_code == "B":
                kwargs["budget_amount"] = field_info.replace(",", "")
            elif line_code == "R":
                kwargs["tax_schedule_info"] = field_info
            else:
                raise ValueError(f"Unknown line code: {line_code}")

        new_cat = cls(**kwargs)
        new_cat.set_parent(new_parent)
        return new_cat


ListOrDictOfCategories = Union[List[Category], Dict[str, Category]]


def create_categories_from_hierarchy(hierarchy: str) -> Category:
    """Create a Category instance from a QIF hierarchy string. Returns the
    lowest category of the hierarchy.
    """
    categories = hierarchy.split(":")
    root_category = Category(name=categories[0])
    current_category = root_category

    if len(categories) > 1:
        for category in categories[1:]:
            current_category = current_category.add_child(category)
    return current_category


def add_categories_to_container(
    new_category: Category,
    categories: ListOrDictOfCategories,
) -> ListOrDictOfCategories:
    """Add ``new_category`` to ``categories`` after first creating necessary
    hierarchy.

    If ``new_category`` already exists somewhere in ``categories``, it will
    be merged with the existing category.

    Parameters
    ----------
    new_category : Category
        The new category to be added to the iterable of categories.
    categories : List[Category] or Dict[str, Category]
        Iterable containing categories. If in a dict, Category objects are the
        values.

    Returns
    -------
    categories : List[Category] or Dict[str, Category]
        The new iterable, now containing ``new_category``
    """
    # Add categories in hierarchy
    categories_is_dict = isinstance(categories, dict)
    if new_category.hierarchy != new_category.name or new_category.children:
        # Get the root category in the chain
        root = new_category.traverse_up()[-1]

        iterator = categories.values() if categories_is_dict else categories

        # Check if the root category already exists in the categories container
        for category in iterator:
            category: Category  # Type hint
            success = category.merge(root)
            if success:
                break
        else:
            # If the category doesn't exist, add it to the categories container
            if categories_is_dict:
                categories[root.name] = root
            else:
                categories.append(root)
    else:
        if categories_is_dict:
            categories[new_category.name] = new_category
        else:
            categories.append(new_category)

    return categories



import re
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from dateutil import parser
from pydantic import ValidationError, parse_obj_as

from quiffen.core.base import Field

ZERO_SEPARATED_DATE = re.compile(
    r"^(\d{2}|\d{4}|[a-zA-Z]+)0(\d{2}|[a-zA-Z]+)0(\d{2}|\d{4})$",
)
INVALID_AMOUNT_CHARACTERS = re.compile(
    r"[^\d\.-]",
)


def parse_date(date_string: str, day_first: bool = False) -> datetime:
    """Parse a string date of an unknown format and return a datetime object.

    Parameters
    ----------
    date_string : str
        String containing date found in QIF file
    day_first : bool, default=False
        Whether the day comes first in the date (e.g. UK date) or after the
        month (e.g. US date)

    Returns
    -------
    datetime.datetime
        datetime object with the date data from the ``date_string`` parameter.

    Raises
    -------
    ValueError
        If the date cannot be parsed.
    """

    # QIF files sometimes use ' ' instead of a 0 or a ' instead of a /
    date_string = date_string.replace(" ", "0")
    date_string = date_string.replace("'", "/")

    # QIF files allow some really strange date formats, such as
    # %d0%B0%Y (e.g. 0100202022 for 2022-02-01)
    # This extends also to month-first dates. The following regex  checks
    # for this and converts it to a date string that can be parsed by
    # dateutil.parser
    date_search = ZERO_SEPARATED_DATE.search(date_string)

    if date_search:
        date_parts = date_search.groups()
        date_string = " ".join(date_parts)

    return parser.parse(date_string, dayfirst=day_first)


def parse_line_code_and_field_info(field: str) -> Tuple[str, str]:
    """Parse a QIF field into a line code and field info."""
    field = field.replace("\n", "")

    if not field:
        return "", ""
    line_code = field[0]

    if len(field) > 1:
        field_info = field[1:]
    else:
        field_info = ""

    return line_code, field_info


def add_custom_field_to_object_dict(
    field: str,  # Takes in the whole field to allow multi-character line codes
    custom_fields: List[Field],
    object_dict: Dict[str, Any],
) -> Tuple[Dict[str, Any], bool]:
    """Add custom QIF fields to a dict representing a Quiffen object based on
    the QIF file line.

    Returns
    -------
    object_dict : dict
        The object dict with the custom field added.
    found : bool
        Whether the custom field was found.
    """
    for custom_field in custom_fields:
        if field.startswith(custom_field.line_code):
            try:
                object_dict[custom_field.attr] = parse_obj_as(
                    custom_field.type,
                    field[len(custom_field.line_code) :],
                )
                return object_dict, True
            except ValidationError:
                return object_dict, False

    return object_dict, False


def convert_custom_fields_to_qif_string(
    custom_fields: List[Field],
    obj: Any,
) -> str:
    """Convert custom fields to a QIF string."""
    qif = ""
    for custom_field in custom_fields:
        if (attr := getattr(obj, custom_field.attr)) is not None:
            qif += f"{custom_field.line_code}{attr}\n"

    return qif


def apply_csv_formatting_to_scalar(
    obj: Any,
    date_format: Optional[str] = "%Y-%m-%d",
    stringify: bool = False,
) -> Any:
    """Apply CSV-friendly formatting to a scalar value"""
    if isinstance(obj, (datetime, date)) and date_format:
        return obj.strftime(date_format)
    elif isinstance(obj, Enum):
        return str(obj.value)
    elif isinstance(obj, Decimal):
        if obj % 1:
            return float(obj)
        return int(obj)
    elif stringify:
        return str(obj)

    return obj


def apply_csv_formatting_to_container(
    obj: Union[List[Any], Dict[Any, Any]],
    date_format: Optional[str] = "%Y-%m-%d",
) -> Union[List[Any], Dict[Any, Any], Any]:
    """Recursively apply CSV-friendly formatting to a container"""
    if isinstance(obj, list):
        return [apply_csv_formatting_to_container(item, date_format) for item in obj]
    elif isinstance(obj, dict):
        return {
            apply_csv_formatting_to_scalar(
                key,
                date_format,
                stringify=True,
            ): apply_csv_formatting_to_container(value, date_format)
            for key, value in obj.items()
        }
    else:
        return apply_csv_formatting_to_scalar(obj, date_format)


def parse_decimal(value: Union[str, Decimal]) -> Decimal:
    """Parse a decimal from a string, removing non-numeric characters."""
    if isinstance(value, Decimal):
        return value
    return Decimal(INVALID_AMOUNT_CHARACTERS.sub("", value))



# Note: this file is named class_type.py rather than class.py as class is a
# reserved word in Python.
from __future__ import annotations

from typing import Any, Dict, List, Optional

from quiffen import utils
from quiffen.core.base import BaseModel, Field
from quiffen.core.category import Category


class Class(BaseModel):
    """
    A class used to represent a QIF Class.

    Parameters
    ----------
    name : str
        The name of the class.
    desc : str, default=None
        The description of the class.
    """

    name: str
    desc: Optional[str] = None
    categories: List[Category] = []

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __eq__(self, other) -> bool:
        if not isinstance(other, Class):
            return False
        return self.name == other.name

    def __str__(self) -> str:
        res = f"Class:\n\tName: {self.name}"
        if self.desc:
            res += f"\n\tDescription: {self.desc}"
        res += f"\n\tCategories: {len(self.categories)}"
        return res

    def add_category(self, new_category: Category) -> None:
        """Add a category to the class."""
        for category in self.categories:
            if category.merge(new_category):
                return

        self.categories.append(new_category)

    def merge(self, other: Class) -> None:
        """Merge another class' categories into this one. Name is not
        merged, and desc is only merged if this class has no desc.
        """
        self.desc = self.desc or other.desc
        for category in other.categories:
            self.add_category(category)

    def to_qif(self) -> str:
        """Return a QIF-formatted string of this class."""
        qif = "!Type:Class\n"
        qif += f"N{self.name}\n"
        if self.desc:
            qif += f"D{self.desc}\n"

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        return qif

    @classmethod
    def from_list(cls, lst: List[str]) -> Class:
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the QIF class.
        """
        kwargs: Dict[str, Any] = {}
        for field in lst:
            line_code, field_info = utils.parse_line_code_and_field_info(field)
            if not line_code:
                continue

            # Check if current line is a custom field
            kwargs, found = utils.add_custom_field_to_object_dict(
                field=field,
                custom_fields=cls._get_custom_fields(),
                object_dict=kwargs,
            )
            if found:
                continue

            if line_code == "N":
                kwargs["name"] = field_info
            elif line_code == "D":
                kwargs["desc"] = field_info
            else:
                raise ValueError(f"Unknown line code: {line_code}")

        return cls(**kwargs)



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import Dict, List, Optional, Union

from quiffen import utils
from quiffen.core.base import BaseModel, Field
from quiffen.core.category import Category
from quiffen.core.class_type import Class


class Split(BaseModel):
    """
    A class used to represent a split in a transaction.


    Examples
    --------
    Adding Splits to a transaction to show that there were two categories that represent
    the transaction.

    >>> import quiffen
    >>> from datetime import datetime
    >>> tr = quiffen.Transaction(date=datetime.now(), amount=150.60)
    >>> beauty = quiffen.Category('Beauty')
    >>> electrical = quiffen.Category('Electrical')
    >>> splt1 = quiffen.Split(amount=50, category=beauty)
    >>> splt2 = quiffen.Split(amount=100.60, category=electrical)
    >>> tr.add_split(splt1)
    >>> tr.add_split(splt2)
    >>> print(tr)
    Transaction:
        Date: 2021-07-05 10:59:02.456190
        Amount: 150.6
        Splits: 2 total split(s)
    >>> print(tr.splits)
    [
        Split(
            amount=50,
            category=Category(name='Beauty', expense=True, hierarchy='Beauty'),
        ),
        Split(
            amount=100.6,
            category=Category(name='Electrical', expense=True, hierarchy='Electrical'),
        ),
    ]
    >>> tr.remove_splits(amount=50)
    [...]
    >>> print(tr.splits)
    [
        Split(
            amount=100.6,
            category=Category(name='Electrical', expense=True, hierarchy='Electrical'),
        ),
    ]
    """

    date: Optional[datetime] = None
    amount: Optional[Decimal] = None
    memo: Optional[str] = None
    cleared: Optional[str] = None
    category: Optional[Category] = None
    to_account: Optional[str] = None
    check_number: Optional[Union[int, str]] = None
    percent: Optional[Decimal] = None
    payee_address: Optional[str] = None

    __CUSTOM_FIELDS: List[Field] = []  # type: ignore

    def __str__(self) -> str:
        properties = ""
        for object_property, value in self.dict().items():
            if value:
                if object_property == "category":
                    properties += f'\n\t\tCategory: {value["name"]}'
                else:
                    properties += (
                        f"\n\t\t"
                        f'{object_property.replace("_", " ").strip().title()}'
                        f": {value}"
                    )

        return "\n\tSplit:" + properties

    def to_qif(
        self,
        date_format: str = "%Y-%m-%d",
        classes: Optional[Dict[str, Class]] = None,
    ) -> str:
        """Returns a QIF string representation of the split."""
        if classes is None:
            classes = {}

        qif = "S"

        if self.category:
            parent_class = None
            for cls in classes.values():
                for category in cls.categories:
                    if category.find_child(self.category.name):
                        parent_class = cls
                        break

            if self.category.hierarchy is not None:
                qif += self.category.hierarchy

            if parent_class:
                qif += f"/{parent_class.name}"

        qif += "\n"

        if self.date:
            qif += f"D{self.date.strftime(date_format)}\n"
        if self.amount:
            qif += f"${self.amount}\n"
        if self.memo:
            qif += f"E{self.memo}\n"
        if self.cleared:
            qif += f"C{self.cleared}\n"
        if self.to_account:
            qif += f"L[{self.to_account}]\n"
        if self.check_number:
            qif += f"N{self.check_number}\n"
        if self.percent:
            qif += f"%{self.percent}%\n"
        if self.payee_address:
            qif += f"A{self.payee_address}\n"

        qif += utils.convert_custom_fields_to_qif_string(
            self._get_custom_fields(),
            self,
        )

        return qif

    @classmethod
    def from_list(cls, lst: List[str]) -> Split:
        raise RuntimeError("Splits can only be created in the context of a transaction")



"""
Quiffen is a Python package for parsing QIF (Quicken Interchange Format) files.

The package allows users to both read QIF files and interact with the contents,
and also to create a QIF structure and then output to either a QIF file, a CSV
of transaction data or a pandas DataFrame.
"""
from quiffen.core.account import Account, AccountType
from quiffen.core.base import Field
from quiffen.core.category import (
    Category,
    CategoryType,
    add_categories_to_container,
    create_categories_from_hierarchy,
)
from quiffen.core.class_type import Class
from quiffen.core.investment import Investment
from quiffen.core.qif import ParserException, Qif, QifDataType
from quiffen.core.security import Security
from quiffen.core.split import Split
from quiffen.core.transaction import Transaction, TransactionLike, TransactionList
